# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""WebVid dataset."""

import io
import os
import pathlib
import subprocess
from typing import List

from absl import logging
from etils import epath
import tensorflow_datasets.public_api as tfds

_FFMPEG_TIMEOUT_SECONDS = 20 * 60  # 20 minutes.
_MAX_SECONDS_PER_VIDEO = 10  # Max allowed length for videos.
# Max allowed size of texts. We assume max 64 words for an English average word
# length of 5. It is important to set a limit to this to avoid stragglers in
# t5x processing.
_MAX_TEXT_LENGTH = 5 * 64
_IMG_SIZE = (360, 640)  # height, width.

_TRAIN_CAPTIONS_GLOB: str = 'results_*_train.csv'
_VAL_CAPTIONS_GLOB: str = 'results_*_val.csv'

_CAPTION_KEY: str = 'name'
_PAGE_DIR_KEY: str = 'page_dir'
_VIDEO_ID_KEY: str = 'videoid'
_URL_KEY: str = 'contentUrl'

_NEW_VIDEO_ID_KEY: str = 'new_video_id'
_VIDEO_PATH_KEY: str = 'video_path'

_FPS = 30
_MAX_FRAMES = _MAX_SECONDS_PER_VIDEO * _FPS
_ENCODING_FORMAT = 'jpeg'


def _split_mjpeg_stream(s: bytes) -> List[bytes]:
  """Split a concatenated string of JPEG images into individual images."""
  # Jpeg files start with ffd8 and end with ffd9.
  start = b'\xff\xd8'
  end = b'\xff\xd9'
  if not (s.startswith(start) and s.endswith(end)):
    raise ValueError('Stream does not start and end correctly.')
  # Split by the start delimiter.
  parts = s.split(start)
  # The stream begins with the start delimiter so the first part should be empty
  assert not parts[0]
  # split() removes the delimiter, so add it back in.
  parts = [start + p for p in parts[1:]]
  # Check that each part starts and ends with the correct delimiters,
  # and check that the delimiters never occur elsewhere
  ok = all(
      (
          p.startswith(start)
          and p.endswith(end)
          and p.count(start) == p.count(end) == 1
      )
      for p in parts
  )
  if not ok:
    raise ValueError('Parts do not start and end correctly.')
  return parts


def _ffmpeg_decode_diskless(
    path,
    extra_ffmpeg_input_args,
    extra_ffmpeg_output_args,
    num_target_frames,
    timeout=None,
):
  """Reads a video from a file, without using the disk.

  Args:
    path: File to read the video from.
    extra_ffmpeg_input_args: Additional input arguments / flags for ffmpeg.
    extra_ffmpeg_output_args: Additional output arguments / flags for ffmpeg.
    num_target_frames: If not None, how many frames to return at most.
    timeout: If not None, maximum amount of seconds to use when running ffmpeg.

  Returns:
    Video frames encoded with encoding_format.
  """
  data = epath.Path(path).read_bytes()
  if not data:
    raise ValueError('Empty file.')

  # FFMPEG args: input from pipe.
  args = ['-i', 'pipe:0']
  args.extend(extra_ffmpeg_input_args)
  # output to pipe.
  args += ['-f', 'image2pipe']
  args.extend(extra_ffmpeg_output_args)

  # Extract the specified number of frames (or all frames if num_frames == 0).
  assert num_target_frames is None or (
      isinstance(num_target_frames, int) and num_target_frames >= 0
  )
  if num_target_frames:
    args += ['-frames:v', f'{num_target_frames}']

  # Output to pipe.
  args += ['-']
  output = tfds.core.utils.image_utils.ffmpeg_run(
      args=args, stdin=data, timeout=timeout
  )

  jpegs = _split_mjpeg_stream(output)
  if num_target_frames:
    jpegs = jpegs[:num_target_frames]
  return jpegs


class Builder(tfds.core.GeneratorBasedBuilder):
  """The WebVid dataset builder."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  Follow the download instructions in https://m-bain.github.io/webvid-dataset/
  to get the data. Place the csv files and the video directories in
  `manual_dir/webvid`, such that mp4 files are placed in
  `manual_dir/webvid/*/*_*/*.mp4`.

  First directory typically being an arbitrary part directory (for sharded
  downloading), second directory is the page directory (two numbers around
  underscore), inside of which there is one or more mp4 files.
  """

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  _homepage: str = 'https://m-bain.github.io/webvid-dataset/'
  _description: str = """\
        WebVid is a large-scale dataset of short videos 
        with textual descriptions sourced from the web. 
        The videos are diverse and rich in their content.

        WebVid-10M contains:

        10.7M video-caption pairs.
        52K total video hours.
        """
  _citation: str = """\
  @misc{bain2021frozen,
        title={Frozen in Time: A Joint Video and Image Encoder for End-to-End Retrieval},
        author={Max Bain and Arsha Nagrani and Gül Varol and Andrew Zisserman},
        year={2021},
        eprint={2104.00650},
        archivePrefix={arXiv},
        primaryClass={cs.CV}
  }
  """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    video_shape = (None,) + _IMG_SIZE + (3,)
    features = {
        'video': tfds.features.Video(
            video_shape, encoding_format=_ENCODING_FORMAT
        ),
        'id': tfds.features.Text(),
        'caption': tfds.features.Text(),
        'url': tfds.features.Text(),
    }

    return tfds.core.DatasetInfo(
        builder=self,
        description=self._description,
        features=tfds.features.FeaturesDict(features),
        supervised_keys=None,
        homepage=self._homepage,
        citation=self._citation,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    def get_captions_path(glob):
      paths = list(epath.Path(dl_manager.manual_dir).glob(glob))
      if len(paths) != 1:
        raise ValueError(
            f'Invalid amount of glob results in {paths} for'
            f' {dl_manager.manual_dir}/{glob}.'
        )
      return paths[0]

    return {
        'train': self._generate_examples(
            captions_csv_path=get_captions_path(_TRAIN_CAPTIONS_GLOB),
            image_base_path=dl_manager.manual_dir,
        ),
        'validation': self._generate_examples(
            captions_csv_path=get_captions_path(_VAL_CAPTIONS_GLOB),
            image_base_path=dl_manager.manual_dir,
        ),
    }

  def _generate_examples(self, captions_csv_path, image_base_path):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam
    pd = tfds.core.lazy_imports.pandas

    ns = 'ProcessExample'
    counter = beam.metrics.Metrics.counter
    total_counter = counter(ns, 'Total')
    process_error_counter = counter(ns, 'Processing error')
    empty_video_counter = counter(ns, 'Empty video')
    timeout_counter = counter(ns, 'Timed out')
    success_counter = counter(ns, 'Success')
    text_cropped_counter = counter(ns, 'Text cropped')

    distribution = beam.metrics.Metrics.distribution
    initial_caption_len_dist = distribution(ns, 'Original caption length')
    final_caption_len_dist = distribution(ns, 'Final caption length')
    frame_count_dist = distribution(ns, 'FrameCount')

    def _load_csv(path):
      # Read all columns as strings, to avoid buggily converting ids with
      # leading zeros into ints.
      return pd.read_csv(
          io.StringIO(epath.Path(path).read_text()),
          sep=',',
          encoding='utf-8',
          dtype='str',
      )

    def _process_example(data_row):
      """Load each video."""
      total_counter.inc()

      file_path = data_row[_VIDEO_PATH_KEY]
      new_video_id = data_row[_NEW_VIDEO_ID_KEY]
      caption = data_row[_CAPTION_KEY]
      url = data_row[_URL_KEY]

      # Read out videos and skip those that don't exist or fail to process.
      height, width = _IMG_SIZE
      extra_input_args = [
          # transform to fps and scale+crop (we must set fps because the input
          # videos don't have consistent fps).
          '-vf',
          f'fps={_FPS},scale={width}:{height}:force_original_aspect_ratio=increase,crop={width}:{height}',  # pylint: disable=line-too-long
          '-sws_flags',
          'lanczos+full_chroma_int+full_chroma_inp+accurate_rnd',
      ]
      extra_output_args = ['-vcodec', 'mjpeg', '-qmin', '1', '-q:v', '2']

      try:
        logging.info('Processing video %s', file_path)

        frames = _ffmpeg_decode_diskless(
            str(file_path),
            extra_ffmpeg_input_args=extra_input_args,
            extra_ffmpeg_output_args=extra_output_args,
            num_target_frames=_MAX_FRAMES,
            timeout=_FFMPEG_TIMEOUT_SECONDS,
        )
      except subprocess.TimeoutExpired as e:
        timeout_counter.inc()
        logging.warning(
            'Timed out while processing %s with exception: %s', file_path, e
        )
        return
      except Exception:  # pylint: disable=broad-except
        process_error_counter.inc()
        logging.exception('Failed to process video %s.', file_path)
        return

      if frames is None:
        empty_video_counter.inc()
        logging.warning('Empty video %s', file_path)
        return

      frame_count_dist.update(len(frames))
      logging.info(
          'Successfully processed %s with original frame count %d',
          file_path,
          len(frames),
      )

      initial_caption_len_dist.update(len(caption))
      if len(caption) > _MAX_TEXT_LENGTH:
        text_cropped_counter.inc()

      features = {
          'id': new_video_id,
          'video': frames[:_MAX_FRAMES],
          'caption': caption[:_MAX_TEXT_LENGTH],
          'url': url,
      }

      final_caption_len_dist.update(len(features['caption']))
      success_counter.inc()
      yield new_video_id, features

    # Get list of videos in file system.
    files = epath.Path(image_base_path).glob(os.path.join('*', '*_*', '*.mp4'))
    df_files = pd.DataFrame(list(files), columns=[_VIDEO_PATH_KEY])
    logging.info('df_files.shape %s', df_files.shape)

    # Read info CSV.
    df_info = _load_csv(captions_csv_path)
    logging.info('df_info.shape %s', df_info.shape)

    def path_to_id(path):
      return os.path.splitext('__'.join(pathlib.Path(path).parts[-2:]))[0]

    # Create base names to join
    df_info[_NEW_VIDEO_ID_KEY] = (
        df_info[_PAGE_DIR_KEY] + '__' + df_info[_VIDEO_ID_KEY]
    )
    df_files[_NEW_VIDEO_ID_KEY] = df_files[_VIDEO_PATH_KEY].map(path_to_id)

    # Merge the two
    df = df_files.merge(df_info, on=_NEW_VIDEO_ID_KEY)

    logging.info('df.shape %s', df.shape)
    logging.info('Number of rows %s', df.shape)
    df = df.to_dict('records')

    return beam.Create(df) | beam.FlatMap(_process_example)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""penguins dataset."""

from __future__ import annotations

import csv
import dataclasses
import textwrap
import typing
from typing import Any, Mapping, Optional, Union

import numpy as np
import tensorflow_datasets.core.utils as type_utils
import tensorflow_datasets.public_api as tfds

_PENGUINS_PATH = (
    'https://storage.googleapis.com/download.tensorflow.org/'
    'data/palmer_penguins/'
)

if typing.TYPE_CHECKING:
  FeatureType = Union[type_utils.TfdsDType, tfds.core.features.FeatureConnector]
else:
  FeatureType = Any


@dataclasses.dataclass
class PenguinConfig(tfds.core.BuilderConfig):
  """Palmer Penguins dataset builder config."""

  # Basename of the file hosting the data.
  file_name: str = ''
  # FeatureDict fields.
  features: Mapping[str, FeatureType] = dataclasses.field(default_factory=dict)
  # Which of the features to use as the label. Packs other features into a
  # single feature for use with `supervised_keys`.
  label: Optional[str] = None
  # What to do with OOV data; field_name -> value.
  cleanup: Optional[Mapping[str, Any]] = None


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for penguins dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      # pytype: disable=wrong-keyword-args
      PenguinConfig(
          name='processed',
          file_name='penguins_processed.csv',
          features={
              'species': tfds.features.ClassLabel(num_classes=3),
              'culmen_length_mm': np.float32,
              'culmen_depth_mm': np.float32,
              'flipper_length_mm': np.float32,
              'body_mass_g': np.float32,
          },
          description=textwrap.dedent(
              """\
            `penguins/processed` is a drop-in replacement for the `iris`
            dataset. It contains 4 normalised numerical features presented as a
            single tensor, no missing values and the class label (species) is
            presented as an integer (n = 334).
          """
          ),
          label='species',
      ),
      PenguinConfig(
          name='simple',
          file_name='penguins_size.csv',
          description=textwrap.dedent(
              """\
            `penguins/simple` has been processed from the raw dataset, with
            simplified class labels derived from text fields, missing values
            marked as NaN/NA and retains only 7 significant features (n = 344).
          """
          ),
          features={
              'species': tfds.features.ClassLabel(
                  names=['Adelie', 'Chinstrap', 'Gentoo']
              ),
              'island': tfds.features.ClassLabel(
                  names=['Biscoe', 'Dream', 'Torgersen']
              ),
              'culmen_length_mm': np.float32,
              'culmen_depth_mm': np.float32,
              'flipper_length_mm': np.float32,
              'body_mass_g': np.float32,
              'sex': tfds.features.ClassLabel(names=['FEMALE', 'MALE', 'NA']),
          },
          label='species',
          cleanup={
              'culmen_length_mm': 'NaN',
              'culmen_depth_mm': 'NaN',
              'flipper_length_mm': 'NaN',
              'body_mass_g': 'NaN',
              'sex': 'NA',
          },
      ),
      PenguinConfig(
          name='raw',
          file_name='penguins_lter.csv',
          description=textwrap.dedent(
              """\
            `penguins/raw` is the original, unprocessed copy from @allisonhorst,
            containing all 17 features, presented either as numeric types or as
            raw text (n = 344).
          """
          ),
          features={
              'studyName': tfds.features.Text(),
              'Sample Number': np.int32,
              'Species': tfds.features.Text(),
              'Region': tfds.features.Text(),
              'Island': tfds.features.Text(),
              'Stage': tfds.features.Text(),
              'Individual ID': tfds.features.Text(),
              'Clutch Completion': tfds.features.Text(),
              'Date Egg': tfds.features.Text(),
              'Culmen Length (mm)': np.float32,
              'Culmen Depth (mm)': np.float32,
              'Flipper Length (mm)': np.float32,
              'Body Mass (g)': np.float32,
              'Sex': tfds.features.Text(),
              'Delta 15 N (o/oo)': np.float32,
              'Delta 13 C (o/oo)': np.float32,
              'Comments': tfds.features.Text(),
          },
          label='Species',
          cleanup={
              'Culmen Length (mm)': 'NaN',
              'Culmen Depth (mm)': 'NaN',
              'Flipper Length (mm)': 'NaN',
              'Body Mass (g)': 'NaN',
              'Delta 15 N (o/oo)': 'NaN',
              'Delta 13 C (o/oo)': 'NaN',
          },
      ),
      # pytype: enable=wrong-keyword-args
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    supervised_keys = None
    features = self.builder_config.features
    supervised_features = features.copy()
    label_name = self.builder_config.label

    if self.builder_config.name == 'processed':
      label_feature = supervised_features.pop(label_name, None)
      supervised_keys = ('features', label_name)
      features = {
          label_name: label_feature,
          'features': tfds.features.Tensor(
              shape=(len(supervised_features),),
              dtype=next(iter(supervised_features.values())),
          ),
      }
    elif self.builder_config.name == 'simple':
      supervised_feature_names = {
          key: key for key in supervised_features.keys()
      }
      supervised_keys = (supervised_feature_names, label_name)

    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(features),
        supervised_keys=supervised_keys,
        homepage='https://allisonhorst.github.io/palmerpenguins/',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download(_PENGUINS_PATH + self.builder_config.file_name)
    return {'train': self._generate_examples(path)}

  def _generate_examples(self, path):
    """Yields examples."""
    with path.open() as f:  # pytype: disable=attribute-error  # gen-stub-imports
      for i, row in enumerate(csv.DictReader(f)):
        row = {f: self._clean_up(f, v) for f, v in row.items()}

        # Pack features if requested.
        if self.builder_config.name == 'processed':
          label_name = self.builder_config.label
          label = row.pop(label_name, None)
          row = list(row.values())
          yield i, {'features': row, label_name: label}
        else:
          yield i, row

  def _clean_up(self, field, value):
    """Applies field-level pre-processing, if needed."""
    if not self.builder_config.cleanup:
      return value
    if field not in self.builder_config.cleanup:
      return value

    feature_type = self.builder_config.features[field]
    if feature_type == np.float32:
      # Field is a float. If it won't parse, clean it up.
      try:
        return float(value)
      except ValueError:
        return self.builder_config.cleanup[field]
    elif isinstance(feature_type, tfds.features.ClassLabel):
      # Field is a class. If it's OOV, clean it up.
      if value not in feature_type.names:
        return self.builder_config.cleanup[field]

    return value



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""imagenet_pi dataset."""

import io
import os
from etils import epath
import numpy as np

from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
import tensorflow_datasets.public_api as tfds

NUM_MODEL_ANNOTATORS = 16


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for imagenet_pi dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  manual_dir should contain two files: ILSVRC2012_img_train.tar and
  ILSVRC2012_img_val.tar.
  You need to register on http://www.image-net.org/download-images in order
  to get the link to download the dataset.
  """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    names_file = imagenet_common.label_names_file()
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'clean_label': tfds.features.ClassLabel(names_file=names_file),
            'annotator_labels': tfds.features.Tensor(
                shape=(NUM_MODEL_ANNOTATORS,), dtype=np.int64
            ),
            'annotator_confidences': tfds.features.Tensor(
                shape=(NUM_MODEL_ANNOTATORS,), dtype=np.float32
            ),
            'file_name': tfds.features.Text(),  # Eg: 'n15075141_54.JPEG'.
        }),
        supervised_keys=('image', 'annotator_labels'),
        homepage='https://github.com/google-research-datasets/imagenet_pi/',
    )

  def _get_values_from_csv(self, input_csv, to_type=int):
    """Gets the annotator labels from the csv lines list."""
    # The input_csv contains lines such as
    # n02097047_2079.JPEG,165,196,196,196,196,196,196,196.
    values_dict = {}
    for cur_line in input_csv:
      elems = cur_line.split(',')
      filename = elems[0]
      values = [to_type(elem) for elem in elems[1:]]
      assert len(values) == NUM_MODEL_ANNOTATORS
      values_dict[filename] = values
    return values_dict

  def _assertion_error_msg(self, missing_item: str, path: str) -> str:
    return (
        f'ImageNet-PI requires manual download of the {missing_item}. '
        f'Please download them and place them into: {path}'
    )

  def _get_train_annotator_labels(self, dl_manager):
    labels_path = dl_manager.manual_dir / 'labels' / 'train.csv'
    labels_path = epath.Path(labels_path)
    if not labels_path.exists():
      raise AssertionError(
          self._assertion_error_msg(
              'train annotator labels', os.fspath(labels_path)
          )
      )
    return self._get_values_from_csv(labels_path.read_text().splitlines())

  def _get_validation_annotator_labels(self, dl_manager):
    labels_path = dl_manager.manual_dir / 'labels' / 'validation.csv'
    labels_path = epath.Path(labels_path)
    if not labels_path.exists():
      raise AssertionError(
          self._assertion_error_msg(
              'validation annotator labels', os.fspath(labels_path)
          )
      )
    return self._get_values_from_csv(labels_path.read_text().splitlines())

  def _get_train_annotator_confidences(self, dl_manager):
    confidences_path = dl_manager.manual_dir / 'confidences' / 'train.csv'
    confidences_path = epath.Path(confidences_path)
    if not confidences_path.exists():
      raise AssertionError(
          self._assertion_error_msg(
              'train annotator confidences', os.fspath(confidences_path)
          )
      )
    return self._get_values_from_csv(
        confidences_path.read_text().splitlines(), to_type=float
    )

  def _get_validation_annotator_confidences(self, dl_manager):
    confidences_path = dl_manager.manual_dir / 'confidences' / 'validation.csv'
    confidences_path = epath.Path(confidences_path)
    if not confidences_path.exists():
      raise AssertionError(
          self._assertion_error_msg(
              'validation annotator confidences', os.fspath(confidences_path)
          )
      )
    return self._get_values_from_csv(
        confidences_path.read_text().splitlines(), to_type=float
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    train_path = dl_manager.manual_dir / 'ILSVRC2012_img_train.tar'
    val_path = dl_manager.manual_dir / 'ILSVRC2012_img_val.tar'

    if not train_path.exists() or not val_path.exists():
      raise AssertionError(
          'ImageNet-PI requires manual download of the ImageNet2012 data. '
          'Please download the train and val set and place them into:'
          '{}, {}'.format(train_path, val_path)
      )

    return {
        'train': self._generate_examples(
            archive=dl_manager.iter_archive(train_path),
            annotator_labels=self._get_train_annotator_labels(dl_manager),
            annotator_confidences=self._get_train_annotator_confidences(
                dl_manager
            ),
            validation_labels=None,
        ),
        'validation': self._generate_examples(
            archive=dl_manager.iter_archive(val_path),
            annotator_labels=self._get_validation_annotator_labels(dl_manager),
            annotator_confidences=self._get_validation_annotator_confidences(
                dl_manager
            ),
            validation_labels=imagenet_common.get_validation_labels(val_path),
        ),
    }

  def _generate_examples(
      self,
      archive,
      annotator_labels,
      annotator_confidences,
      validation_labels=None,
  ):
    # Validation split.
    if validation_labels:
      for fname, fobj in archive:
        assert fname in validation_labels
        assert fname in annotator_labels
        assert fname in annotator_confidences
        record = {
            'file_name': fname,
            'image': fobj,
            'clean_label': validation_labels[fname],
            'annotator_labels': annotator_labels[fname],
            'annotator_confidences': annotator_confidences[fname],
        }
        yield fname, record

    # Training split. Main archive contains archives names after a synset noun.
    # Each sub-archive contains pictures associated to that synset.
    for fname, fobj in archive:
      assert fname.endswith('.tar')
      label = fname[:-4]  # fname is something like 'n01632458.tar'
      fobj_mem = io.BytesIO(fobj.read())
      for image_fname, image in tfds.download.iter_archive(
          fobj_mem, tfds.download.ExtractMethod.TAR_STREAM
      ):
        record = {
            'file_name': image_fname,
            'image': image,
            'clean_label': label,
            'annotator_labels': annotator_labels[image_fname],
            'annotator_confidences': annotator_confidences[image_fname],
        }
        yield image_fname, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QuALITY dataset."""
import dataclasses
import json

import numpy as np
import tensorflow_datasets.public_api as tfds

_DOWNLOAD_URL = (
    'https://github.com/nyu-mll/quality/raw/main/data/QuALITY.v0.9.zip'
)

# Fields that are straight text copies from raw example to processed example.
_ONE2ONE_FIELDS = (
    'article',
    'article_id',
    'set_unique_id',
    'writer_id',
    'source',
    'title',
    'topic',
    'url',
    'writer_id',
)


@dataclasses.dataclass
class QualityConfig(tfds.core.BuilderConfig):
  stripped: bool = False


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for quality dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  BUILDER_CONFIGS = [
      QualityConfig(
          name='raw', description='Raw with HTML.', stripped=False
      ),  # default
      QualityConfig(
          name='stripped', description='Stripped of HTML.', stripped=True
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    # Mirror format of RACE dataset as much as possible.
    features_dict = {
        'article': tfds.features.Text(),
        # The sequence lengths of Sequence fields should match.
        'questions': tfds.features.Sequence(tfds.features.Text()),
        'question_ids': tfds.features.Sequence(tfds.features.Text()),
        # 4 options per question, similar to RACE
        'options': tfds.features.Sequence(
            tfds.features.Sequence(tfds.features.Text())
        ),
        'gold_labels': tfds.features.Sequence(np.int32),  # 0, 1, 2, 3
        'writer_labels': tfds.features.Sequence(np.int32),  # 0, 1, 2, 3
        'difficults': tfds.features.Sequence(np.bool_),
    }
    features_dict.update({k: tfds.features.Text() for k in _ONE2ONE_FIELDS})
    return self.dataset_info_from_configs(
        # Note: some fields are left out.
        features=tfds.features.FeaturesDict(features_dict),
        supervised_keys=None,
        homepage='https://github.com/nyu-mll/quality',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(_DOWNLOAD_URL)

    if self.builder_config.stripped:
      return {
          'train': self._generate_examples(
              path / 'QuALITY.v0.9.htmlstripped.train', 'train'
          ),
          'test': self._generate_examples(
              path / 'QuALITY.v0.9.htmlstripped.test', 'test'
          ),
          'dev': self._generate_examples(
              path / 'QuALITY.v0.9.htmlstripped.dev', 'dev'
          ),
      }
    else:
      return {
          'train': self._generate_examples(
              path / 'QuALITY.v0.9.train', 'train'
          ),
          'test': self._generate_examples(path / 'QuALITY.v0.9.test', 'test'),
          'dev': self._generate_examples(path / 'QuALITY.v0.9.dev', 'dev'),
      }

  def _generate_examples(self, path, split: str):
    """Yields examples."""
    for line in path.open():
      j = json.loads(line)
      fields = {k: j[k] for k in _ONE2ONE_FIELDS}
      fields.update({
          'questions': [q['question'] for q in j['questions']],
          'question_ids': [q['question_unique_id'] for q in j['questions']],
          'difficults': [q['difficult'] for q in j['questions']],
          'options': [q['options'] for q in j['questions']],
      })
      if split in ('train', 'dev'):
        fields.update({
            'gold_labels': [q['gold_label'] for q in j['questions']],
            'writer_labels': [q['writer_label'] for q in j['questions']],
        })
      else:
        fields.update({
            'gold_labels': [],
            'writer_labels': [],
        })
      yield j['set_unique_id'], fields



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Imagenet LT datasets."""

import io
import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
import tensorflow_datasets.public_api as tfds

_TRAIN_SUBSET = (
    'https://drive.google.com/uc?export=download&'
    'id=1Sl1cwy6Dei1I1BMS1YKjI35fkaR1UA_7'
)

_VAL_SUBSET = (
    'https://drive.google.com/uc?export=download&'
    'id=1AjYczW4khrQrwPIygXUkHF-Qv4QGEsCF'
)


class Builder(tfds.core.GeneratorBasedBuilder):
  """Long-tailed version of the ImageNet2012 dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  manual_dir should contain two files: ILSVRC2012_img_train.tar and
  ILSVRC2012_img_val.tar.
  You need to register on http://www.image-net.org/download-images in order
  to get the link to download the dataset.
  """

  def _info(self):
    """Define the dataset info."""
    names_file = imagenet_common.label_names_file()
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'label': tfds.features.ClassLabel(names_file=names_file),
            'file_name': tfds.features.Text(),  # Eg: 'n15075141_54.JPEG'
        }),
        supervised_keys=('image', 'label'),
        homepage='https://github.com/zhmiao/OpenLongTailRecognition-OLTR',
    )

  def _postprocess_subset_list(self, subset):
    """Postprocess the subset list."""

    # The original subset contains lines such as
    # "train/n01440764/n01440764_190.JPEG 0".
    # Postprocessing removes the redundant info and only keeps the image name,
    # i.e. n01440764_190.JPEG.
    post_subset = [
        os.path.basename(cur_item.split(' ')[0]) for cur_item in subset
    ]
    return post_subset

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    train_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_train.tar')
    val_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_val.tar')

    # We don't import the original test split, as it doesn't include labels.
    # These were never publicly released.
    if not tf.io.gfile.exists(train_path) or not tf.io.gfile.exists(val_path):
      raise AssertionError(
          'ImageNet-LT requires manual download of the ImageNet2012 data. '
          'Please download the train and val set and place them into:'
          '{}, {}'.format(train_path, val_path)
      )

    # Download and load subset file.
    downloaded_dirs = dl_manager.download({
        'train': _TRAIN_SUBSET,
        'validation': _VAL_SUBSET,
    })

    with tf.io.gfile.GFile(downloaded_dirs['train']) as fp:
      train_subset = fp.read().splitlines()
    train_subset = self._postprocess_subset_list(train_subset)

    with tf.io.gfile.GFile(downloaded_dirs['validation']) as fp:
      val_subset = fp.read().splitlines()
    val_subset = self._postprocess_subset_list(val_subset)

    return {
        'train': self._generate_examples(
            archive=dl_manager.iter_archive(train_path), subset=train_subset
        ),
        'validation': self._generate_examples(
            archive=dl_manager.iter_archive(train_path), subset=val_subset
        ),
        'test': self._generate_examples(
            archive=dl_manager.iter_archive(val_path),
            validation_labels=imagenet_common.get_validation_labels(val_path),
        ),
    }

  def _generate_examples(self, archive, subset=None, validation_labels=None):
    """Yields examples."""
    # Test split in ImageNet-LT is the validation split in the original ImageNet
    if validation_labels:
      for key, example in imagenet_common.generate_examples_validation(
          archive, validation_labels
      ):
        yield key, example

    # Training and validation split. Main archive contains archives names after
    # a synset noun.
    # Each sub-archive contains pictures associated to that synset.
    for fname, fobj in archive:
      label = fname[:-4]  # fname is something like 'n01632458.tar'
      # TODO(b/117643231): in py3, the following lines trigger tarfile module
      # to call `fobj.seekable()`, which Gfile doesn't have. We should find an
      # alternative, as this loads ~150MB in RAM.
      fobj_mem = io.BytesIO(fobj.read())
      for image_fname, image in tfds.download.iter_archive(
          fobj_mem, tfds.download.ExtractMethod.TAR_STREAM
      ):  # pytype: disable=wrong-arg-types  # gen-stub-imports
        if subset is None or image_fname in subset:  # filtering using subset.
          record = {
              'file_name': image_fname,
              'image': image,
              'label': label,
          }
          yield image_fname, record
